/**
 * TCPConnection.java
 *
 Copyright (C) 2001 Geometrek	and 2009 by R. Russell Kinter
 
 *  This file is part of DeepMatrixIP9.
 *
 *  DeepMatrixIP9 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  DeepMatrix IP9 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * This class implements a Connection over a TCP stream. It encapsultes a Socket
 * and the In/Outputstreams used for communication.
 * It gets or sends Messages and is used in both the client and
 * the server program.
 *
 * @see Message
 * @version 1.2
 */
public class TCPConnection extends Connection {
    private Socket ConSocket = null;
    private DataInputStream in;
    private DataOutputStream out;
    private boolean alive;
    private int buffer = 2048;
    private MessageBuffer msgBuffer;
  /** 
   * New 11-12-05 An added MsgBuffer for handling chat text ControlMessages
   * corresponds to the method sendMessage3
   */
    private MessageBuffer msgBuffer3;
    
/**
 * New 11-12-05 this class implements a send buffer for either RealTimeMessages, or chat
 * text ControlMessages. It has been altered from original design in that the original was
 * for RealTimeMessages only and checked for duplicates. This checking code has been removed
 * because the number of RealTimeMessages sent has been reduced 75% in the client. This inner
 * class is now simply a buffer.
 */
    private class MessageBuffer implements Runnable {
        private boolean kill = false;
        private Vector buffer;
        private Thread runner;
   /** 
   * New 11-12-05 yn is a boolean used when the Thread is shut down. Since there are two message buffers,
   * the first one to be shut down is told to wait 1 millisecond. The last one is notified in the standard
   * way. yn simply tells the stop method which is meant to be first and last. See stop method below.
   */
        private boolean yn;
        
        MessageBuffer( boolean tf ){
            yn  = tf;
            buffer = new Vector(10,10);
            runner = new Thread( this, "msgBuffer" );
            runner.start();
        }
        
        public void run(){
            Message msg;
            while( Thread.currentThread() == runner ){
                     try{
                         if(!buffer.isEmpty()){
                        synchronized( buffer ){
                             msg =(Message) buffer.firstElement();
                              buffer.removeElementAt(0);
                             }
                         sendRawMessage( msg );
                         }
                    else {
                         synchronized( buffer ){
                            if(kill==false){
                             buffer.wait();
                           }
                             else {
                         runner = null;
                            }
                          }
                        }
                    }
                catch(InterruptedException e){}
                catch(Exception e){
                    System.out.println("msgBuffer Error "+e);
                }
            }
        }
        
        public void put( Message msg ){
            synchronized( buffer ){
                buffer.addElement( msg );
                buffer.notify();
            }
        }

   /** 
   * New 11-12-05 yieldRunner is called to hopefully "brake" the threads of this inner class
   * for the priority messages send by sendMessage2 of the main TCPConnection class.
   */    
        public void yieldRunner() throws InterruptedException {
          synchronized( buffer ){
                buffer.wait(100);
            }
       }

  /** 
   * New 11-12-05 Since there are two message buffers, the first one to be shut down is told to wait 1 millisecond. 
   * The last one is notified in the standard way. boolean yn simply tells the stop method which is meant to be first and last.
   * See yn global variable above.
   */

        public void stop() throws InterruptedException {
           
            synchronized( buffer ){
              kill = true;
              buffer.removeAllElements();
              buffer.notify(); 
               }
            }
        }
    
    public TCPConnection() {
        alive = false;
    }
    
    public TCPConnection( int buffer ){
        this.buffer = buffer;
        alive = false;
    }
    
    public TCPConnection( Socket socks ) throws IOException, InterruptedException {
        open( socks );
    }
    
    public TCPConnection( Socket socks, int buffer ) throws IOException, InterruptedException {
        this.buffer = buffer;
        open( socks );
    }
    
    public TCPConnection( String host, int port ) throws IOException, InterruptedException {
        open( host, port );
    }
    
    public TCPConnection( String host, int port, int buffer ) throws IOException, InterruptedException {
        this.buffer = buffer;
        open( host, port );
    }

    /** 
   * New 11-12-05 setSoTimeout is now called and set for 16 seconds. This directly relates to the new ControlMessage
   * HRTBT 650 sent to the server every seven seconds from the client. If a client doesn't contact the server in
   * someway ( Any Message will work besides HRTBT, but what if the user is neither moving or chatting? )
   * the server will boot the client.
   */ 

    public void open( Socket socks ) throws IOException, InterruptedException {
        close();
        ConSocket = socks;
        in = new DataInputStream( new BufferedInputStream( ConSocket.getInputStream(), buffer) );
        out = new DataOutputStream( new BufferedOutputStream( ConSocket.getOutputStream(), buffer) );
        alive = true;
        msgBuffer = new MessageBuffer(true);
        msgBuffer3 = new MessageBuffer(false);
    }
    
    public void open( String host, int port ) throws IOException, InterruptedException {
        open( new Socket( host, port ));
    }
    
    public void close() throws InterruptedException {
        if( msgBuffer != null ){
            msgBuffer.stop();}
         if( msgBuffer3 != null ){
            msgBuffer3.stop();}
        alive = false;
        try {
            if( ConSocket != null ) {
                in.close();
                out.close();
                ConSocket.close();
            }
        }
        catch(IOException e) {
        in = null;
        out = null;
        ConSocket = null;
         }
      msgBuffer = null;
      msgBuffer3 = null;
    }
    
    protected void sendRawMessage( Message msg ) throws IOException {
        synchronized( out ){
            out.writeInt( msg.getType());
            msg.write(out);
            out.flush();
        }
    }

 /**
 * New 11-12-05 sendMessage exclusively RealTimeMessages
 */ 
    
    public void sendMessage( Message msg ) throws IOException {
            msgBuffer.put( msg );
    }

 /**
 * New 11-12-05 sendMessage2 will tell the msgBuffers to yield because it has priority for 
 * sending update information. Admittedly, it is a little doubtful how effective the yield method is 
 * with threads.
 */ 
  
public void sendMessage2( Message msg ) throws IOException, InterruptedException { 
       msgBuffer.yieldRunner();
       msgBuffer3.yieldRunner();
       sendRawMessage( msg );
    }

  /**
 * New 11-12-05 sendMessage3 exclusively chat text ControlMessages
 */ 
    
  public void sendMessage3( Message msg ) throws IOException {
        msgBuffer3.put( msg );
  }


    
    public Message getMessage() throws IOException {
        Message msg;
        synchronized( in ){
            int type = in.readInt();
            switch( type ){
                case ControlMessage.TYPE :
                    msg = new ControlMessage( in );
                    break;
                case RealTimeMessage.TYPE :
                    msg = new RealTimeMessage( in );
                    break;
				case RealTimeSave.TYPE :
                    msg = new RealTimeSave( in );
                    break;
				case WhoWhatWhere.TYPE :
                    msg = new WhoWhatWhere( in );
                    break;
				case AdminControl.TYPE :
                    msg = new AdminControl( in );
                    break;
				case RoomControl.TYPE :
                    msg = new RoomControl( in );
                    break;
                case ObjectEnvelope.TYPE :
					msg = new ObjectEnvelope( in );
                    break;
				case TriggerEnvelope.TYPE :
                    msg = new TriggerEnvelope( in );
                    break;
				case ControlEvent.TYPE :
                    msg = new ControlEvent( in );
                    break;
                    default :
                        throw new UnknownMessageException("unknown message type "+type);
            }
        }
        return msg;
    }

   public void exchange( ObjectInputStream in, ObjectOutputStream out ) throws Exception {
   }

    
  public boolean isAlive() { return alive; }
  
  public void setSoTimeout( int iSTO ) throws SocketException {
    ConSocket.setSoTimeout( iSTO );
  } 
   
    public InetAddress getInetAddress(){
        if(isAlive())
            return ConSocket.getInetAddress();
        else
            return null;
    }

    public  void startMsgBuffer(){
   }

    public int getPort(){
        if(isAlive())
            return ConSocket.getPort();
        else
            return 0;
    }

  public void setShutDown() throws IOException, InterruptedException {
      setSoTimeout(2000);
      ConSocket.shutdownInput();
      ConSocket.shutdownOutput();
    }

}

