/*
  HttpServletConnection.java
 
  Copyright (C) 2001 Geometrek
 
 *  This file is part of DeepMatrixIP9.
 *
 *  DeepMatrixIP9 is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  DeepMatrix IP9 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package matrix.server;

import matrix.*;
import java.io.*;
import java.net.*;
import java.util.*;

/** This class provides the Server side HTTP tunneling implementation of
 * a connection. It stores messages to and from the client and sends them
 * to the client, whenever a request is made from the MatrixServlet servlet.
 * @author gr
 */
public class HttpServletConnection extends Connection {
    
    /** stores the Internet address of the connecting client
    */    
    protected InetAddress address;
    /** list of incoming and outgoing messages
    */    
    protected Vector input = new Vector(), output = new Vector();
    
    /** Creates new HttpServletConnection
     * @param address the Internet address of the connecting client
    */
    public HttpServletConnection( InetAddress address ) {
        this.address = address;
        alive = true;
    }
   
    public HttpServletConnection() {
    }
 
    // finalize method clears message buffers
    protected void finalize() throws Throwable {
        input.removeAllElements();
        output.removeAllElements();
        super.finalize();
    }
    
    /** this method has no implementation and use in the HttpServletConnection.
    * It throws an exception, if called.
    * @param host method should not be used
    * @param port method should not be used
    * @throws IOException will always be thrown
    */    
    public void open( String host, int port ) throws IOException {
        throw new IOException("Call to open is not supported by HttpServletConnection !");
    }
    
    /** closes the connection
    */
    public void close(){
        if( isAlive()){
            // maybe some internal protocol message ?
        }
        alive = false;
        // input.removeAllElements();
        // output.removeAllElements();
    }
 
    /** sends a message over the connection
    * @param msg message to send
    * @throws IOException thrown if anything goes wrong
    */
    public void sendMessage( Message msg ) throws IOException {
        if( !isAlive()) {
            throw new IOException("Connection not connected !");
        }       
        output.addElement( msg.copy() );
    }

    public void sendMessage2( Message msg ) throws IOException, InterruptedException {
               if( !isAlive()) {
            throw new IOException("Connection not connected !");
        }
        output.addElement( msg.copy() );
   }
    
    public void sendMessage3( Message msg ) throws IOException {
                 if( !isAlive()) {
            throw new IOException("Connection not connected !");
        }       
        output.addElement( msg.copy() );
       }

    /** reads a message from the connection
    * @throws IOException thrown if anything goes wrong
    * @return the last message read
    */
    public Message getMessage() throws IOException {
        if( !isAlive()) {
            throw new IOException("Connection not connected !");
        }       
        Message msg;
        try {
            synchronized( input ){
                if( input.size() == 0 ){
                    input.wait();
                }
            }
            msg = (Message) input.firstElement();
            input.removeElementAt(0);            
        }
        catch( InterruptedException e ){
            throw new InterruptedIOException("Timeout");
        }
        return msg;
    }        
    
    /** exchanges messages by reading and writing data from and to an Http Request
    * vs. Reply object. It is only passed object streams to use.
    * @param in object input stream to read incoming messages from
    * @param out object output stream to write outgoing messages to
    */    
   public void exchange( ObjectInputStream in, ObjectOutputStream out ) throws Exception {
        /*if( !isAlive())
            throw new IOException("Connection not connected !");*/
        Message[] msgs;
        Exception error = null;
        try {
            msgs = (Message[]) in.readObject();       
            for( int i = 0; i < msgs.length; i++ ){
                input.addElement( msgs[i] );
            }
            if( msgs.length > 0 ){
                synchronized( input ){
                    input.notifyAll();
                }            
            }
            msgs = null;
        }
        catch(Exception e){
            error = e;
        }        
        synchronized( output ){ 
            msgs = new Message[ output.size() ];
            output.copyInto( msgs );
            try{
                out.writeObject( msgs );
            }
            catch(IOException e){
                error = e;
            }
            output.removeAllElements();
        }      
        if( error != null )
            throw error;
    }

    public void startMsgBuffer() {
    }

	public boolean isAlive() { return alive; }

     public void setSoTimeout( int iSTO ) throws SocketException {
     }  

    /** returns the InetAddress of the host the connection is currently connected to.
    * @return Address or null, if not connected
    */

    public InetAddress getInetAddress(){
        if( isAlive() ){
            return address;
        }
        return null;
    }
    
    /** allways returns 80, because this is a HttpServlet connection, even if
    * this is not really useful or correct.
    * @return the port or 0, if not connected
    */
    public int getPort(){
        if( isAlive()){
            return 80;
        }
        return 0;
    }

   public void setShutDown() throws IOException, InterruptedException {
   }

}
